# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Calculate and draw a dragon curve using Python turtle graphics."""

from __future__ import annotations

import dataclasses
import sys
import turtle
import typing

import click

from . import calc
from . import defs
from . import util


if typing.TYPE_CHECKING:
    from typing import Final


@dataclasses.dataclass
class ConfigHolder:
    """Hold a `Config` object."""

    cfg: defs.Config | None = None
    """The `Config` object stashed by the main function."""


def extract_cfg(ctx: click.Context) -> defs.Config:
    """Extract the `Config` object that the main function built."""
    cfg_hold: Final = ctx.find_object(ConfigHolder)
    if cfg_hold is None:
        sys.exit("Internal error: no click config holder object")

    cfg: Final = cfg_hold.cfg
    if cfg is None:
        sys.exit("Internal error: no config in the click config holder")

    return cfg


def arg_features(_ctx: click.Context, _self: click.Parameter, value: bool) -> bool:  # noqa: FBT001
    """Display program features information and exit."""
    if not value:
        return value

    print("Features: " + " ".join(f"{name}={value}" for name, value in defs.FEATURES.items()))
    sys.exit(0)


@click.command(name="calc")
@click.argument("generations", type=int, required=True, nargs=-1)
@click.pass_context
def cmd_calc(ctx: click.Context, *, generations: list[int]) -> None:
    """Calculate the turn at the specified generation."""
    cfg: Final = extract_cfg(ctx)
    for generation in generations:
        print(f"{generation}\t{calc.turn(cfg, generation)}")


@click.command(name="run")
@click.option(
    "--turtle-delay",
    type=int,
    default=1,
    help="the delay between successive steps of the turtle",
)
@click.option(
    "--turtle-step",
    type=int,
    default=5,
    help="the length of the turtle's forward movement",
)
@click.pass_context
def cmd_run(ctx: click.Context, *, turtle_delay: int, turtle_step: int) -> None:
    """Draw the dragon curve using Python turtle graphics."""
    if turtle_delay < 0:
        sys.exit("The --turtle-delay value must be non-negative")
    if turtle_step < 1:
        sys.exit("The --turtle-step value must be positive")

    cfg: Final = extract_cfg(ctx)
    current = 1
    turtle.hideturtle()
    turtle.delay(turtle_delay)
    while True:
        turtle.forward(turtle_step)
        if calc.turn(cfg, current) == defs.Turn.LEFT:
            turtle.left(90)
        else:
            turtle.right(90)

        current += 1


@click.command(name="turns")
@click.argument("generation", type=int, required=True)
@click.pass_context
def cmd_turns(ctx: click.Context, *, generation: int) -> None:
    """Display the successive turns."""
    cfg: Final = extract_cfg(ctx)
    for current in range(1, generation + 1):
        print(calc.turn(cfg, current))


@click.group(name="turtle-dragon")
@click.option(
    "--features",
    is_flag=True,
    is_eager=True,
    callback=arg_features,
    help="display program features information and exit",
)
@click.option(
    "--quiet",
    "-q",
    is_flag=True,
    help="quiet operation; only display warning and error messages",
)
@click.option("--verbose", "-v", is_flag=True, help="verbose operation; display diagnostic output")
@click.pass_context
def main(ctx: click.Context, *, features: bool, quiet: bool, verbose: bool) -> None:
    """Do something, or do something else, who cares anyway."""
    if features:
        sys.exit("Internal error: how did we get to main() with features=True?")

    ctx.ensure_object(ConfigHolder)
    ctx.obj.cfg = defs.Config(cache={}, log=util.build_logger(quiet=quiet, verbose=verbose))


main.add_command(cmd_calc)
main.add_command(cmd_run)
main.add_command(cmd_turns)


if __name__ == "__main__":
    main()
