# SPDX-FileCopyrightText: Peter Pentchev <roam@ringlet.net>
# SPDX-License-Identifier: BSD-2-Clause
"""Calculate the turns of a dragon curve."""

from __future__ import annotations

import typing

from . import defs


if typing.TYPE_CHECKING:
    from typing import Final


CACHE_MAX: Final = 131071
"""The largest generation value that we are willing to cache."""


def turn(cfg: defs.Config, generation: int) -> defs.Turn:
    """Calculate successive turns."""

    def update(value: defs.Turn) -> defs.Turn:
        """Update the cache if necessary."""
        if generation in cfg.cache:
            raise RuntimeError(repr((cfg, generation, value)))

        if generation <= CACHE_MAX:
            cfg.cache[generation] = value

        return value

    if generation < 1:
        raise ValueError(generation)

    match cfg.cache.get(generation):
        case None:
            pass

        case cached:
            cfg.log.debug(
                "turn: %(gen)d: cached: %(cached)s",
                {"gen": generation, "cached": cached},
            )
            return cached

    count = 0
    pow2 = generation
    next_pow2 = pow2 & (pow2 - 1)
    while (pow2 & next_pow2) != 0:
        count += 1
        pow2 = next_pow2
        next_pow2 = pow2 & (pow2 - 1)

    if count == 0:
        cfg.log.debug("turn: %(gen)d: power of two, easy", {"gen": generation})
        return update(defs.Turn.LEFT)

    cfg.log.debug("turn: %(gen)d: pow2 %(pow2)d", {"gen": generation, "pow2": pow2})
    return update(turn(cfg, (pow2 << 1) - generation).converse())
